# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.dispatch import Signal


message_received = Signal(["message_info"])
"""收到微信推送消息"""

message_handled = Signal(["message_info", "reply"])
"""微信推送消息处理成功(不包括成功)"""

message_error = Signal(["message_info", "exc"])
"""微信推送消息处理异常"""



# -*- coding: utf-8 -*-

"""微信消息处理器"""

from __future__ import unicode_literals

from contextlib import contextmanager
from functools import wraps
import logging
import time

from django.core.cache import cache
from django.http import response
from django.utils.datastructures import MultiValueDictKeyError
from django.views import View
import six
from wechatpy import replies
from wechatpy.events import BaseEvent
from wechatpy.exceptions import InvalidSignatureException
from wechatpy.utils import check_signature
import xmltodict

from . import settings, signals
from .exceptions import BadMessageRequest, MessageHandleError

__all__ = ("handler", "handle_subscribe_events", "Handler", "message_handler",
           "message_rule")


class Handler(View):
    def dispatch(self, request):
        """
        :type request: wechat_django.requests.WeChatMessageRequest
        """
        if not request.wechat.app.abilities.interactable:
            return response.HttpResponseNotFound()
        try:
            self._verify(request)
            resp = super(Handler, self).dispatch(request)
            if not isinstance(resp, response.HttpResponseNotFound):
                self.log(logging.DEBUG, "receive a message")
            return resp
        except Exception as e:
            return self._handle_exception(request, e)

    def get(self, request):
        return request.GET["echostr"]

    def post(self, request):
        from .models import WeChatMessageInfo
        from .sites.wechat import patch_request

        request = patch_request(request, cls=WeChatMessageInfo)
        message_info = request.wechat
        signals.message_received.send(request.wechat.app.staticname,
                                      message_info=message_info)
        try:
            reply = self._handle(message_info)
            signals.message_handled.send(request.wechat.app.staticname,
                                         message_info=message_info,
                                         reply=reply)
            if reply:
                xml = reply.render()
                if request.wechat.app.crypto:
                    xml = request.wechat.app.crypto.encrypt_message(
                        xml, request.GET["nonce"], request.GET["timestamp"])
                return response.HttpResponse(xml, content_type="text/xml")
        except Exception as exc:
            signals.message_error.send(request.wechat.app.staticname,
                                       message_info=message_info, exc=exc)
            raise
        return ""

    def _verify(self, request):
        """检验请求"""
        try:
            timestamp = int(request.GET["timestamp"])
        except ValueError:
            raise BadMessageRequest("invalid timestamp")
        sign = request.GET["signature"]
        nonce = request.GET["nonce"]

        time_diff = int(timestamp) - time.time()
        # 检查timestamp
        if abs(time_diff) > settings.MESSAGETIMEOFFSET:
            raise BadMessageRequest("invalid time")

        # 防重放检查及签名检查
        with self._no_repeat_nonces(sign, nonce, time_diff):
            check_signature(
                request.wechat.app.token,
                sign,
                timestamp,
                nonce
            )

    @contextmanager
    def _no_repeat_nonces(self, sign, nonce, time_diff):
        """nonce防重放"""
        nonce_key = "wx:m:n:{0}".format(sign)
        if settings.MESSAGENOREPEATNONCE:
            if cache.get(nonce_key) == nonce:
                raise BadMessageRequest("repeat nonce string")
            try:
                yield
                expires = settings.MESSAGETIMEOFFSET + time_diff
                cache.set(nonce_key, nonce, expires)
            finally:
                pass
        else:
            yield

    def _handle(self, message_info):
        """处理消息"""
        from .models import MessageHandler, MessageLog

        handlers = MessageHandler.matches(message_info)
        if not handlers:
            return None

        handler = handlers[0]
        reply = handler.reply(message_info)
        if handler.log_message or message_info.app.log_message:
            MessageLog.from_message_info(message_info)
        if not reply or isinstance(reply, replies.EmptyReply):
            return None
        return reply

    def _handle_exception(self, request, exc):
        if isinstance(exc, MultiValueDictKeyError):
            self.log(logging.WARNING, "bad request args", exc_info=True)
            return response.HttpResponseBadRequest()
        elif isinstance(exc, BadMessageRequest):
            self.log(logging.WARNING, "bad request", exc_info=True)
            return response.HttpResponseBadRequest()
        elif isinstance(exc, InvalidSignatureException):
            self.log(logging.WARNING, "invalid signature", exc_info=True)
            return response.HttpResponseBadRequest()
        elif isinstance(exc, xmltodict.expat.ExpatError):
            self.log(logging.WARNING, "deserialize message failed", exc_info=True)
            return response.HttpResponseBadRequest()
        elif isinstance(exc, MessageHandleError):
            self.log(logging.WARNING, "handle message failed", exc_info=True)
            return ""
        else:
            self.log(logging.ERROR, "an unexcepted error occurred", exc_info=True)
            return ""

    _log = None

    @property
    def log(self):
        if not self._log:
            from .models import MessageHandler
            self._log = MessageHandler.handlerlog(self.request)
        return self._log


def message_handler(names_or_func=None):
    """
    自定义回复业务需加装该装饰器
    被装饰的自定义业务接收一个``wechat_django.models.WeChatMessageInfo``对象
    并且返回一个``wechatpy.replies.BaseReply``对象

    :param names_or_func: 允许使用该message_handler的appname 不填所有均允许
    :type names_or_func: str or list or tuple or callable

        @message_handler
        def custom_business(message):
            user = message.user
            # ...
            return TextReply("hello")

        @message_handler(("app_a", "app_b"))
        def app_ab_only_business(message):
            # ...
    """
    return _decorator("message_handler", names_or_func)


def message_rule(names_or_func=None):
    """
    自定义规则需加装该装饰器
    接收一个`wechat_django.models.WeChatMessageInfo`对象
    返回一个bool值,真值则表示规则符合

        @message_rule
        def custom_rule(message_info):
            user = message.user
            # ...
            return True

        @message_rule(("app_a", "app_b"))
        def app_ab_only_rule(message):
            # ...
    """
    return _decorator("message_rule", names_or_func)


def _decorator(property, names_or_func):
    def decorator(view_func):
        @wraps(view_func)
        def decorated_view(message):
            return view_func(message)
        setattr(decorated_view, property, names or True)

        return decorated_view

    if isinstance(names_or_func, six.text_type):
        names = [names_or_func]
    elif callable(names_or_func):
        names = None
        return decorator(names_or_func)

    return decorator


def handle_subscribe_events(sender, message_info, **kwargs):
    """处理关注,取关"""
    message = message_info.message
    if isinstance(message, BaseEvent):
        # 关注事件
        if message.event in ("subscribe", "subscribe_scan"):
            message_info.local_user.subscribe = True
            message_info.local_user.subscribe_time = time.time()
            message_info.local_user.save()

        # 取关事件
        if message.event == "unsubscribe":
            message_info.local_user.subscribe = False
            message_info.local_user.save()


handler = Handler.as_view()



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import unicode_literals

from wechat_django.constants import WeChatSNSScope
from .authentication import (
    WeChatOAuthAuthentication, WeChatOAuthSessionAuthentication)
from .client import WeChatOAuthClient
from .mixins import WeChatOAuthViewMixin
from .permissions import WeChatAuthenticated
from .views import wechat_auth, WeChatOAuthView



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

try:
    from rest_framework.authentication import BaseAuthentication
except ImportError:
    BaseAuthentication = object
from six.moves.urllib.parse import parse_qsl, urlparse
from wechatpy import WeChatOAuthException


class WeChatOAuthAuthentication(BaseAuthentication):
    """
    兼容drf的网页授权认证,仅可通过code授权
    """
    def authenticate(self, request):
        wechat = request.wechat
        code = self._get_params(request, "code")
        if wechat.openid:
            return wechat.user, wechat.openid
        elif code:
            try:
                user = self._auth(wechat.app, code, wechat.scope)
                wechat._openid = user.openid
                wechat._user = user
                # 用当前url的state替换传入的state
                wechat._state = self._get_params(request, "state", "")
                return wechat.user, wechat.openid
            except WeChatOAuthException:
                err_msg = "auth code failed: {0}".format(dict(
                    info=wechat,
                    code=code
                ))
                wechat.app.logger("oauth").warning(err_msg, exc_info=True)

    def authenticate_header(self, request):
        return 'WOAuth realm="{0}"'.format(request.wechat.app.appid)

    def _auth(self, app, code, scope):
        return app.auth(code, scope)[0]

    def _get_params(self, request, key, default=None):
        """获取url上的参数"""
        if request.is_ajax():
            try:
                referrer = request.META["HTTP_REFERER"]
                query = dict(parse_qsl(urlparse(referrer).query))
                return query.get(key, default)
            except:
                return default
        else:
            return request.GET.get(key, default)


class WeChatOAuthSessionAuthentication(WeChatOAuthAuthentication):
    """
    兼容drf的网页授权认证,包含session设置,一般使用该认证
    """

    def authenticate(self, request):
        session_key = request.wechat.session_key
        request.wechat._openid = request.session.get(session_key)
        rv = super(WeChatOAuthSessionAuthentication, self).authenticate(
            request)
        if rv is not None:
            request.session[session_key] = rv[1]
        return rv



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.core.exceptions import PermissionDenied
from django.views import View


class APIView(View):
    """
    A minimum implementation for drf APIView with permission and
    authentication
    """

    def permission_denied(self, request, message=None):
        """
        If request is not permitted, determine what kind of exception to raise.
        """
        raise PermissionDenied(message)

    def get_authenticators(self):
        """
        Instantiates and returns the list of authenticators that this view can use.
        """
        return [auth() for auth in self.authentication_classes]

    def get_permissions(self):
        """
        Instantiates and returns the list of permissions that this view requires.
        """
        return [permission() for permission in self.permission_classes]

    def perform_authentication(self, request):
        """
        Perform authentication on the incoming request.

        Note that if you override this and simply 'pass', then authentication
        will instead be performed lazily, the first time either
        `request.user` or `request.auth` is accessed.
        """
        for authenticator in self.get_authenticators():
            user_auth_tuple = authenticator.authenticate(request)
            if user_auth_tuple is not None:
                return

    def check_permissions(self, request):
        """
        Check if the request should be permitted.
        Raises an appropriate exception if the request is not permitted.
        """
        for permission in self.get_permissions():
            if not permission.has_permission(request, self):
                self.permission_denied(
                    request, message=getattr(permission, 'message', None)
                )

    def initialize_request(self, request, *args, **kwargs):
        """
        Returns the initial request object.
        """
        return request

    def initial(self, request, *args, **kwargs):
        """
        Runs anything that needs to occur prior to calling the method handler.
        """
        # Ensure that the incoming request is permitted
        self.perform_authentication(request)
        self.check_permissions(request)

    def finalize_response(self, request, response, *args, **kwargs):
        """
        Returns the final response object.
        """
        return response

    def handle_exception(self, exc):
        """
        Handle any exception that occurs, by returning an appropriate response,
        or re-raising the error.
        """
        raise exc

    def dispatch(self, request, *args, **kwargs):
        """
        `.dispatch()` is pretty much the same as Django's regular dispatch,
        but with extra hooks for startup, finalize, and exception handling.
        """
        self.args = args
        self.kwargs = kwargs
        request = self.initialize_request(request, *args, **kwargs)
        self.request = request

        try:
            self.initial(request, *args, **kwargs)
            response = super(APIView, self).dispatch(request, *args, **kwargs)
        except Exception as exc:
            response = self.handle_exception(exc)

        self.response = self.finalize_response(request, response, *args,
                                               **kwargs)
        return self.response

    def setup(self, request, *args, **kwargs):
        """Initialize attributes shared by all view methods."""
        self.request = request
        self.args = args
        self.kwargs = kwargs



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from datetime import timedelta
import time

from django.urls import reverse
from django.utils import timezone
from wechatpy import events, messages

from ..handler import Handler, message_handler
from ..models import (MessageHandler, Reply, Rule, WeChatMessageInfo,
                      WeChatUser)
from ..sites.wechat import patch_request
from .base import WeChatTestCase
from .interceptors import (common_interceptor, wechatapi,
    wechatapi_accesstoken, wechatapi_error)


@message_handler
def debug_handler(message):
    return "success"


@message_handler("test")
def app_only_handler(message):
    return "success"


def forbidden_handler(message):
    return ""


class HandlerTestCase(WeChatTestCase):
    def test_available(self):
        """测试handler有效性"""
        rule = dict(type=Rule.Type.ALL)
        now = timezone.now()
        day = timedelta(days=1)
        handler_not_begin = self._create_handler(rule, starts=now + day)
        handler_ended = self._create_handler(rule, ends=now - day)
        handler_disabled = self._create_handler(rule, enabled=False)
        handler_available = self._create_handler(
            rule, starts=now - day, ends=now + day)

        msg_info = self._msg2info(messages.TextMessage("abc"))
        self.assertFalse(handler_not_begin.is_match(msg_info))
        self.assertFalse(handler_ended.is_match(msg_info))
        self.assertFalse(handler_disabled.is_match(msg_info))
        self.assertTrue(handler_available.is_match(msg_info))

        matches = MessageHandler.matches(msg_info)
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0], handler_available)

    def test_sync(self):
        """测试同步"""
        pass

    def test_subscribe_events(self):
        """测试关注,取关事件"""
        # url = reverse("wechat_django:handler",
        #               kwargs=dict(appname=self.app.name))
        url = "/"
        openid = "test_subscribe_events"
        user = WeChatUser.objects.upsert_by_dict(dict(openid=openid), self.app)
        self.assertIsNone(user.subscribe)

        handler = Handler()

        subscribe_event_text = """
        <xml>
        <ToUserName><![CDATA[ToUser]]></ToUserName>
        <FromUserName><![CDATA[{0}]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[subscribe]]></Event>
        </xml>
        """.format(openid)
        request = self.rf().post(
            url, subscribe_event_text, content_type="text/xml")
        request = patch_request(request, self.app.name)
        self.assertEqual(handler.post(request), "")
        user.refresh_from_db()
        self.assertTrue(user.subscribe)

        unsubscribe_event_text = """
        <xml>
        <ToUserName><![CDATA[ToUser]]></ToUserName>
        <FromUserName><![CDATA[{0}]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[unsubscribe]]></Event>
        </xml>
        """.format(openid)
        request = self.rf().post(
            url, unsubscribe_event_text, content_type="text/xml")
        request = patch_request(request, self.app.name)
        self.assertEqual(handler.post(request), "")
        user.refresh_from_db()
        self.assertEqual(user.subscribe, False)

        subscribe_event_text = """
        <xml>
        <ToUserName><![CDATA[ToUser]]></ToUserName>
        <FromUserName><![CDATA[{0}]]></FromUserName>
        <CreateTime>123456789</CreateTime>
        <MsgType><![CDATA[event]]></MsgType>
        <Event><![CDATA[subscribe]]></Event>
        <EventKey><![CDATA[qrscene_123123]]></EventKey>
        <Ticket><![CDATA[TICKET]]></Ticket>
        </xml>
        """.format(openid)
        request = self.rf().post(
            url, subscribe_event_text, content_type="text/xml")
        request = patch_request(request, self.app.name)
        self.assertEqual(handler.post(request), "")
        user.refresh_from_db()
        self.assertTrue(user.subscribe)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from wechat_django.constants import WeChatSNSScope
from wechat_django.utils.web import auto_response

from .authentication import WeChatOAuthSessionAuthentication


class WeChatOAuthViewMixin(object):

    appname = None
    """
    微信appname 必填
    """

    scope = (WeChatSNSScope.BASE,)
    """
    :type: str or iterable
    微信授权的scope 默认WeChatSNSScope.BASE
    """

    redirect_uri = None
    """
    :type: str or Callable[
        [
            django.http.request.HttpRequest,
            *args,
            **kwargs
        ],
        str
    ]

    未授权时的重定向地址 当未设置response时将自动执行授权
    当ajax请求时默认取referrer 否则取当前地址
    注意 请不要在地址上带有code及state参数 否则可能引发问题
    """

    state = ""
    """
    :type: str or Callable[
        [
            django.http.request.HttpRequest,
            *args,
            **kwargs
        ],
        str
    ]

    授权时需要携带的state
    """

    authentication_classes = (WeChatOAuthSessionAuthentication,)

    def __init__(self, *args, **kwargs):
        super(WeChatOAuthViewMixin, self).__init__(*args, **kwargs)
        assert self.appname, "appname must be set"

    def initialize_request(self, request, *args, **kwargs):
        request = super(WeChatOAuthViewMixin, self).initialize_request(
            request, *args, **kwargs)
        return self._initialize_wechat_request(request, *args, **kwargs)

    def finalize_response(self, request, response, *args, **kwargs):
        response = auto_response(response)
        response = super(WeChatOAuthViewMixin, self).finalize_response(
            request, response, *args, **kwargs)
        return response

    def _initialize_wechat_request(self, request, *args, **kwargs):
        from wechat_django.models import WeChatOAuthInfo
        from wechat_django.sites.wechat import patch_request

        state = self.state(request, *args, **kwargs) if callable(self.state)\
            else self.state
        redirect_uri = self.redirect_uri(request, *args, **kwargs)\
            if callable(self.redirect_uri)\
            else self.redirect_uri
        return patch_request(
            request=request,
            appname=self.appname,
            cls=WeChatOAuthInfo,
            redirect_uri=redirect_uri,
            scope=self.scope,
            state=state
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

try:
    from rest_framework.permissions import BasePermission
except ImportError:
    BasePermission = object


class WeChatAuthenticated(BasePermission):
    def has_permission(self, request, view):
        return hasattr(request, "wechat") and bool(request.wechat.openid)

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json

from django.conf.urls import url
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import response
from django.test import RequestFactory
from django.test.utils import override_settings
from django.utils.http import urlencode
import six
from six.moves.urllib.parse import parse_qsl, urlparse

from ..models import WeChatUser
from ..oauth import (WeChatAuthenticated, wechat_auth, WeChatSNSScope,
                     WeChatOAuthClient, WeChatOAuthView)
from .base import mock, WeChatTestCase
from .interceptors import common_interceptor


def oauth_api(openid, callback=None):
    kwargs = dict(
        netloc=r"(.*\.)?api\.weixin\.qq\.com$",
        path="/sns/oauth2/access_token"
    )

    def mock(url, request):
        headers = {
            "Content-Type": "application/json"
        }
        data = dict(
            access_token="123",
            refresh_token="123",
            expires_in=7200,
            openid=openid
        )
        from httmock import response
        resp = response(200, data, headers)
        if callback:
            callback(url, request, resp)
        return resp

    return common_interceptor(mock, **kwargs)


def snsinfo_api(openid, callback=None, **data):
    kwargs = dict(
        netloc=r"(.*\.)?api\.weixin\.qq\.com$",
        path="/sns/userinfo"
    )

    def mock(url, request):
        headers = {
            "Content-Type": "application/json"
        }
        from httmock import response
        resp = response(200, data, headers)
        if callback:
            callback(url, request, resp)
        return resp

    return common_interceptor(mock, **kwargs)


@wechat_auth("test", state="state")
def test_oauth_view(request, *args):
    user = request.wechat.user
    return dict(
        openid=user.openid,
        id=user.id,
        args=args
    )

urlpatterns = [
    url(r"^test/(.+)$", test_oauth_view)
]


@override_settings(ROOT_URLCONF=__name__)
class OAuthTestCase(WeChatTestCase):
    def test_oauth_api(self):
        """测试oauth接口请求"""
        pass

    def test_redirect_uri(self):
        """测试redirect_uri生成"""
        host = "example.com"
        rf = self.rf(HTTP_HOST=host)
        api = "/app"
        absolute_uri = lambda uri: "http://" + host + uri

        # 测试写入redirect_uri时
        redirect_uri = "https://mp.weixin.qq.com/wiki"
        view_cls = WeChatOAuthView(
            appname=self.app.name, redirect_uri=redirect_uri)
        request = rf.get(api)
        SessionMiddleware().process_request(request)
        view_cls.setup(request)
        request = view_cls.initialize_request(request)
        self.assertEqual(request.wechat.redirect_uri, redirect_uri)

        redirect_uri = "/app"
        view_cls = WeChatOAuthView(
            appname=self.app.name, redirect_uri=redirect_uri)
        request = rf.get(api)
        SessionMiddleware().process_request(request)
        view_cls.setup(request)
        request = view_cls.initialize_request(request)
        self.assertEqual(request.wechat.redirect_uri,
                         absolute_uri(redirect_uri))

        # 测试模板请求redirect_uri
        view_cls = WeChatOAuthView(appname=self.app.name)
        request = rf.get(redirect_uri)
        SessionMiddleware().process_request(request)
        view_cls.setup(request)
        request = view_cls.initialize_request(request)
        self.assertEqual(request.wechat.redirect_uri,
                         absolute_uri(redirect_uri))

        # 测试ajax请求redirect_uri
        view_cls = WeChatOAuthView(appname=self.app.name)
        request = rf.post(api, dict(a=1),
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
            HTTP_REFERRER=absolute_uri(redirect_uri))
        SessionMiddleware().process_request(request)
        view_cls.setup(request)
        request = view_cls.initialize_request(request)
        self.assertEqual(request.wechat.redirect_uri,
                         absolute_uri(redirect_uri))

        # 测试callable redirect_uri
        callable_redirect = lambda request, *args, **kwargs: redirect_uri
        view_cls = WeChatOAuthView(appname=self.app.name,
                                   redirect_uri=callable_redirect)
        request = rf.get(redirect_uri)
        SessionMiddleware().process_request(request)
        view_cls.setup(request)
        request = view_cls.initialize_request(request)
        self.assertEqual(request.wechat.redirect_uri,
                         absolute_uri(redirect_uri))

    def test_unauthorization(self):
        rf = RequestFactory()
        path = "/test"
        redirect_uri = "https://mp.weixin.qq.com/wiki"

        # 默认response
        state = "abc"
        request = rf.get(path)
        handler = self._make_handler(
            request, redirect_uri=redirect_uri,
            scope=WeChatSNSScope.USERINFO, state=state)
        request = handler.initialize_request(request)
        resp = handler.unauthorization_response(request)
        self.assertIsInstance(resp, response.HttpResponseRedirect)
        client = WeChatOAuthClient(self.app)
        self.assertEqual(resp.url, client.authorize_url(
            redirect_uri, WeChatSNSScope.USERINFO, state))

        # 传入callable state
        state = lambda request: request.path
        request = rf.get(path)
        handler = self._make_handler(
            request, redirect_uri=redirect_uri,
            scope=WeChatSNSScope.USERINFO, state=state)
        request = handler.initialize_request(request)
        resp = handler.unauthorization_response(request)
        self.assertIsInstance(resp, response.HttpResponseRedirect)
        client = WeChatOAuthClient(self.app)
        self.assertEqual(resp.url, client.authorize_url(
            redirect_uri, WeChatSNSScope.USERINFO, path))

        # 传入callback response
        resp = response.HttpResponseForbidden()
        request = rf.get(path)
        handler = self._make_handler(request, response=resp)
        request = handler.initialize_request(request)
        self.assertIs(
            resp,
            handler.unauthorization_response(request)
        )

        # 传入普通response
        resp_func = lambda request: response.HttpResponseForbidden(request.path)
        request = rf.get(path)
        handler = self._make_handler(request, response=resp_func)
        request = handler.initialize_request(request)
        resp = handler.unauthorization_response(request)
        self.assertIsInstance(resp, response.HttpResponseForbidden)
        self.assertEqual(resp.content, path.encode())

    def test_user_update(self):
        "测试更新授权数据"
        pass

    def test_request(self):
        "测试请求"
        redirect_uri = "https://mp.weixin.qq.com/wiki"
        url = "/test"
        request = self._create_request(url)

        # 设置了response
        resp = response.HttpResponseForbidden()
        handler = self._make_handler(
            request, redirect_uri=redirect_uri, response=resp)
        self.assertIs(handler.dispatch(request), resp)

        # 未授权
        handler = self._make_handler(request, redirect_uri=redirect_uri)
        resp = handler.dispatch(request)
        self.assertIsInstance(resp, response.HttpResponseRedirect)
        self.assertEqual(resp.url, request.wechat.oauth_uri)

        # 授权
        request = self._create_request(url + "?code=123")
        openid = "456"
        view = lambda request, *args, **kwargs: request.wechat.openid
        handler = self._make_handler(request, view=view)
        with oauth_api(openid):
            resp = handler.dispatch(request)
            self.assertEqual(resp.content, openid.encode())

        # 已授权
        session_key = "sessionid"
        resp = SessionMiddleware().process_response(request, resp)
        request = self._create_request(url + "?code=123")
        request.COOKIES[session_key] = resp.cookies[session_key].value
        resp = handler.dispatch(request)
        self.assertEqual(resp.content, openid.encode())

    def test_decorator(self):
        """测试class based view与decorator是否一致"""
        openid = "test_decorator"
        user = WeChatUser.objects.upsert_by_dict(
            dict(openid=openid), app=self.app)
        session_key = "wechat_{0}_user".format(self.app.name)

        view = lambda request: request
        # 测试appname
        wrapped_view = wechat_auth(self.app.name)(view)
        request = self._create_request("/")
        request.session[session_key] = openid
        self.assertEqual(wrapped_view(request).wechat.appname, self.app.name)

        # 测试scope
        wrapped_view = wechat_auth(
            self.app.name, scope=WeChatSNSScope.USERINFO)(view)
        request = self._create_request("/")
        request.session[session_key] = openid
        self.assertIn(WeChatSNSScope.USERINFO, wrapped_view(request).wechat.scope)

        # 测试redirect_uri
        redirect_uri = "https://www.baidu.com/"
        wrapped_view = wechat_auth(
            self.app.name, redirect_uri=redirect_uri)(view)
        request = self._create_request("/")
        request.session[session_key] = openid
        self.assertEqual(wrapped_view(request).wechat.redirect_uri, redirect_uri)

        # 测试required
        wrapped_view = wechat_auth(self.app.name, required=False)(view)
        request = self._create_request("/")
        self.assertEqual(wrapped_view(request), request)

        # 测试默认required及response
        resp = response.HttpResponse()
        wrapped_view = wechat_auth(self.app.name, response=resp)(view)
        request = self._create_request("/")
        self.assertEqual(wrapped_view(request), resp)

        # 测试state
        state = "abc"
        wrapped_view = wechat_auth(
            self.app.name, state=state)(view)
        request = self._create_request("/")
        request.session[session_key] = openid
        self.assertEqual(wrapped_view(request).wechat.state, state)

        # 测试methods
        request_get = self._create_request("/")
        request_get.session[session_key] = openid
        request_post = self._create_request("/", "post")
        request_post.session[session_key] = openid

        # 测试仅get
        view = lambda request: resp
        wrapped_view = wechat_auth(self.app.name)(view)
        self.assertEqual(wrapped_view(request_get), resp)
        self.assertIsInstance(wrapped_view(request_post),
                              response.HttpResponseNotAllowed)
        
        # 测试仅post
        wrapped_view = wechat_auth(self.app.name, methods=["POST"])(view)
        self.assertIsInstance(wrapped_view(request_get),
                              response.HttpResponseNotAllowed)
        self.assertEqual(wrapped_view(request_post), resp)

    def test_view(self):
        """测试view是否正常"""
        args = "t"
        openid = "666"
        host = "example.com"
        path = "/test/" + args

        resp = self.client.get(path, HTTP_HOST="example.com")
        self.assertEqual(resp.status_code, 302)
        parsed_url = urlparse(resp.url)
        self.assertEqual(parsed_url.netloc, "open.weixin.qq.com")
        query = dict(parse_qsl(parsed_url.query))
        self.assertEqual(query["appid"], self.app.appid)
        self.assertEqual(query["response_type"], "code")
        self.assertEqual(query["scope"], WeChatSNSScope.BASE)
        self.assertEqual(query["state"], "state")
        self.assertEqual(parsed_url.fragment, "wechat_redirect")
        self.assertEqual(query["redirect_uri"], "http://" + host + path)

        with oauth_api(openid):
            resp = self.client.get(path + "?code=123", follow=True,
                HTTP_HOST=host)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content.decode())
        self.assertEqual(data["openid"], openid)
        self.assertEqual(data["args"], [args])
        id = data["id"]
        user = WeChatUser.objects.get(id=id)
        self.assertEqual(user.openid, openid)

    def test_custom_oauth_url(self):
        """测试设置了OAUTH_URL url后 不再向微信请求授权,转向第三方请求授权"""
        with mock.patch.object(WeChatOAuthClient, "authorize_url"):
            new_url = "new_url"
            self.app.configurations["OAUTH_URL"] = new_url
            hasattr(self.app, "_oauth") and delattr(self.app, "_oauth")
            url = self.app.oauth.authorize_url("redirect_url")
            self.assertTrue(url.startswith(new_url))
            delattr(self.app, "_oauth")
            del self.app.configurations["OAUTH_URL"]

    def _make_handler(self, request, appname="", **kwargs):
        view = kwargs.pop("view", lambda request, *args, **kwargs: "")
        appname = appname or self.app.name
        scope = kwargs.get("scope")
        if scope and isinstance(scope, six.text_type):
            kwargs["scope"] = (scope,)
        cls = type(str("WeChatOAuthView"), (WeChatOAuthView,), dict())
        kwargs = cls.prepare_init_kwargs(appname=appname, **kwargs)
        rv = cls(**kwargs)
        rv.setup(request)
        rv.get = view
        return rv

    def _create_request(self, path, method="get"):
        request = getattr(self.rf(), method)(path)
        SessionMiddleware().process_request(request)
        return request



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.utils.http import urlencode
from wechatpy import WeChatOAuth

from wechat_django.constants import WeChatSNSScope


class WeChatOAuthClient(WeChatOAuth):
    OAUTH_URL = "https://open.weixin.qq.com/connect/oauth2/authorize"
    QRCONNECT_URL = "https://open.weixin.qq.com/connect/qrconnect"

    def __init__(self, app):
        if app.configurations.get("OAUTH_URL"):
            self.OAUTH_URL = app.configurations["OAUTH_URL"]
        super(WeChatOAuthClient, self).__init__(app.appid, app.appsecret, "")

    def authorize_url(self, redirect_uri, scope=WeChatSNSScope.BASE, state=""):
        return self.OAUTH_URL + "?" + urlencode(dict(
            appid=self.app_id,
            redirect_uri=redirect_uri,
            response_type="code",
            scope=scope,
            state=state
        )) + "#wechat_redirect"

    def qrconnect_url(self, redirect_uri, state=""):
        return self.QRCONNECT_URL + "?" + urlencode(dict(
            appid=self.app_id,
            redirect_uri=redirect_uri,
            response_type="code",
            scope="snsapi_login",
            state=state
        )) + "#wechat_redirect"



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from functools import wraps

from django.http.response import HttpResponse
from django.shortcuts import redirect
import six
from wechatpy import WeChatOAuthException

from wechat_django.constants import WeChatSNSScope
from wechat_django.rest_framework.views import APIView

from .mixins import WeChatOAuthViewMixin
from .permissions import WeChatAuthenticated


class WeChatOAuthView(WeChatOAuthViewMixin, APIView):

    required = True
    """真值必须授权 否则不授权亦可继续访问(只检查session)"""

    response = None
    """
    :type: django.http.response.HttpResponse or Callable[
        [
            django.http.request.HttpRequest,
            *args,
            **kwargs
        ],
        django.http.response.HttpResponse
    ]
    """

    permission_classes = tuple()

    @classmethod
    def prepare_init_kwargs(cls, **initKwargs):
        # 检查属性正确性
        appname = initKwargs.get("appname") or cls.appname
        assert appname and isinstance(appname, six.text_type),\
            "incorrect appname"

        response = initKwargs.get("response") or cls.response
        assert (
            response is None or callable(response)
            or isinstance(response, HttpResponse)
        ), "incorrect response param"

        scope = initKwargs.get("scope") or cls.scope
        if isinstance(scope, six.text_type):
            scope = (scope,)
        for s in scope:
            assert s in (WeChatSNSScope.BASE, WeChatSNSScope.USERINFO),\
                "incorrect scope"

        # 对于必须授权的请求 在permissions中添加WeChatAuthenticated
        required = initKwargs.get("required")
        if required is None:
            required = cls.required
        # 重新处理base_permissions
        base_permissions = getattr(cls, "permission_classes", tuple())
        base_type = type(base_permissions)
        base_permissions = list(base_permissions)
        if required and WeChatAuthenticated not in base_permissions:
            initKwargs["permission_classes"] = base_type(
                [WeChatAuthenticated] + base_permissions)
        elif not required and WeChatAuthenticated in base_permissions:
            base_permissions.remove(WeChatAuthenticated)
            initKwargs["permission_classes"] = base_type(base_permissions)

        return initKwargs

    @classmethod
    def as_view(cls, **initKwargs):
        initKwargs = cls.prepare_init_kwargs(**initKwargs)
        return super(WeChatOAuthView, cls).as_view(**initKwargs)

    def check_permissions(self, request):
        for permission in self.get_permissions():
            if not permission.has_permission(request, self):
                if isinstance(permission, WeChatAuthenticated):
                    raise WeChatOAuthException(
                        0, getattr(permission, "message", None))
                else:
                    self.permission_denied(
                        request, message=getattr(permission, 'message', None)
                    )

    def handle_exception(self, exc):
        if isinstance(exc, WeChatOAuthException):
            return self.unauthorization_response(
                self.request, *self.args, **self.kwargs)

        return super(WeChatOAuthView, self).handle_exception(exc)

    def unauthorization_response(self, request, *args, **kwargs):
        """未授权的响应"""
        response = self.response
        if response and callable(response):
            response = response(request, *args, **kwargs)
        elif not response:
            oauth_uri = request.wechat.oauth_uri
            response = redirect(oauth_uri, permanent=False)
        return response


def wechat_auth(appname, scope=None, redirect_uri=None, required=True,
                response=None, state="", methods=None):
    """
    微信网页授权
    :param appname: WeChatApp的name
    :param scope: 默认WeChatSNSScope.BASE, 可选WeChatSNSScope.USERINFO
    :type scope: str or iterable
    :param redirect_uri: 未授权时的重定向地址 当未设置response时将自动执行授权
                        当ajax请求时默认取referrer 否则取当前地址
                        注意 请不要在地址上带有code及state参数 否则可能引发问题 
    :type redirect_uri: str or Callable[
        [
            django.http.request.HttpRequest,
            *args,
            **kwargs
        ],
        str
    ]
    :param state: 授权时需要携带的state
    :type state: str
                 or Callable[
                    [
                        django.http.request.HttpRequest,
                        *args,
                        **kwargs
                    ],
                    str
                 ]
    :param required: 真值必须授权 否则不授权亦可继续访问(只检查session)
    :param response: 未授权的返回 接受一个
    :type response: django.http.response.HttpResponse
                    or Callable[
                        [
                            django.http.request.HttpRequest,
                            *args,
                            **kwargs
                        ],
                        django.http.response.HttpResponse
                    ]
    :param methods: 允许的HTTP请求方法,默认仅GET
    :type methods: tuple

    使用示例:

        @wechat_auth("your_app_name")
        def your_view(request, *args, **kwargs):
            # request是一个``wechat_django.requests.WeChatOAuthRequest对象``
            user = request.wechat.user
    """

    methods = methods or ("GET",)

    def decorator(func):
        @wraps(func)
        def view(self, request, *args, **kwargs):
            return func(request, *args, **kwargs)

        attrs = {method.lower(): view for method in methods}
        attrs["appname"] = appname
        View = type(str("WeChatOAuthView"), (WeChatOAuthView,), attrs)

        return View.as_view(appname=appname, scope=scope,
                            redirect_uri=redirect_uri, required=required,
                            response=response, state=state)
    return decorator



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.sessions.middleware import SessionMiddleware
from wechatpy.oauth import WeChatOAuthException

from ..models import WeChatApp, WeChatUser
from wechat_django.oauth import (WeChatAuthenticated, WeChatOAuthClient,
                                 WeChatOAuthSessionAuthentication,
                                 WeChatOAuthView, WeChatSNSScope)
from .base import mock, WeChatTestCase


class OAuthAuthenticationTestCase(WeChatTestCase):
    def test_authentication(self):
        """测试WeChatOAuthSessionAuthentication"""

        path = "/test"
        openid = "openid_test_authentication"

        auth = WeChatOAuthSessionAuthentication()
        user_dict = dict(openid=openid)
        wechat_user = WeChatUser.objects.upsert_by_dict(user_dict, self.app)

        # 未认证用户返回False header返回非空
        request = self._create_request(path)
        view = self._make_view(request)
        request = view.initialize_request(request)
        self.assertIsNone(auth.authenticate(request))
        self.assertTrue(auth.authenticate_header(request))

        # code认证用户返回True 并正确记录session
        code = "123456"
        request = self._create_request(path + "?code=" + code)
        request = view.initialize_request(request)
        with mock.patch.object(WeChatApp, "auth"):
            WeChatApp.auth.return_value = wechat_user, user_dict
            user, authed_openid = auth.authenticate(request)
            self.assertTrue(user.id)
            self.assertEqual(user.openid, openid)
            self.assertEqual(authed_openid, openid)
            self.assertEqual(request.session[request.wechat.session_key],
                             openid)
            self.assertEqual(request.wechat.user.openid, openid)
            self.assertEquals(WeChatApp.auth.call_args[0][0], code)

        # session认证用户返回True
        request = self._create_request(path + "?code=" + code)
        request = view.initialize_request(request)
        request.session[request.wechat.session_key] = openid
        with mock.patch.object(WeChatApp, "auth"):
            user, authed_openid = auth.authenticate(request)
            self.assertEqual(user.openid, openid)
            self.assertEqual(authed_openid, openid)
            self.assertEqual(request.session[request.wechat.session_key],
                             openid)
            self.assertEqual(request.wechat.user.openid, openid)
            # session认证用户不调用code认证代码
            self.assertFalse(WeChatApp.auth.called)

        # 认证异常用户返回False
        # 假code
        request = self._create_request(path + "?code=" + code)
        request = view.initialize_request(request)
        with mock.patch.object(WeChatApp, "auth"):
            WeChatApp.auth.side_effect = WeChatOAuthException(0, "")
            self.assertIsNone(auth.authenticate(request))
            self.assertTrue(auth.authenticate_header(request))

    def test_model_auth(self):
        """测试WeChatApp.auth"""

        code = "code"
        openid = "openid_test_model_auth"

        # base授权
        with mock.patch.object(WeChatOAuthClient, "fetch_access_token"):
            WeChatOAuthClient.fetch_access_token.return_value = dict(
                openid=openid)

            user, user_dict = self.app.auth(code, WeChatSNSScope.BASE)
            self.assertTrue(user.id)
            self.assertEqual(user.openid, openid)
            self.assertEqual(user_dict["openid"], openid)
            WeChatOAuthClient.fetch_access_token.assert_called_with(code)

        # snsinfo授权
        openid = "openid_test_model_auth2"
        nickname = "nickname"
        user_info = dict(openid=openid, nickname=nickname)
        with mock.patch.object(WeChatOAuthClient, "fetch_access_token"),\
             mock.patch.object(WeChatOAuthClient, "get_user_info"):
            WeChatOAuthClient.fetch_access_token.return_value = dict(
                openid=openid)
            WeChatOAuthClient.get_user_info.return_value = user_info

            user, user_dict = self.app.auth(code, WeChatSNSScope.USERINFO)
            self.assertTrue(user.id)
            self.assertEqual(user.openid, openid)
            self.assertEqual(user.nickname, nickname)
            self.assertEqual(user_dict, user_info)
            WeChatOAuthClient.fetch_access_token.assert_called_with(code)

    def test_permissions(self):
        """测试WeChatAuthenticated权限"""
        path = "/test"
        openid = "openid_test_permissions"

        permission = WeChatAuthenticated()
        user_dict = dict(openid=openid)
        WeChatUser.objects.upsert_by_dict(user_dict, self.app)

        # 未认证用户返回False
        request = self._create_request(path)
        view = self._make_view(request)
        request = view.initialize_request(request)
        self.assertFalse(permission.has_permission(request, view))

        # TODO: 怎么测认证失败用户返回False

        # 认证用户返回True
        request = self._create_request(path)
        view = self._make_view(request)
        request = view.initialize_request(request)
        request.session[request.wechat.session_key] = openid
        self.assertTrue(permission.has_permission(request, view))
        self.assertTrue(permission.has_object_permission(request, view, None))

    def _make_handler(self, request, appname="", **kwargs):
        view = kwargs.pop("view", lambda request, *args, **kwargs: "")
        rv = self._make_view(request, appname, **kwargs)
        rv.initialize_request(request)
        rv.get = view
        return rv

    def _make_view(self, request, appname="", **kwargs):
        appname = appname or self.app.name
        rv = WeChatOAuthView(appname=appname, **kwargs)
        rv.setup(request)
        return rv

    def _create_request(self, path):
        request = self.rf().get(path)
        SessionMiddleware().process_request(request)
        return request



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _

from . import settings


class WeChatConfig(AppConfig):
    name = "wechat_django"
    verbose_name = _("WeChat")
    verbose_name_plural = _("WeChat")

    def ready(self):
        import object_tool
        object_tool.ObjectToolConfig.register()

        if settings.PATCHADMINSITE:
            from .sites.admin import patch_admin
            patch_admin()

        # 注册注册关注,取关事件
        from . import signals
        from .handler import handle_subscribe_events

        signals.message_received.connect(handle_subscribe_events)
